/* $Id: rt_prton.c,v 1.3 2008/11/30 17:31:10 kiesling Exp $ */

/*
  This file is part of ctalk.
  Copyright  2005-2008  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"
#include "rtinfo.h"
#include "typeof.h"

extern char ptr_args[512][8192];    /* Declared in rt_stdarg.c. */
extern int ptr_arg_idx;
char fmt_tokens[512][8192];
int fmt_tok_idx;

static char tmparg0buf[0xFFFF];

int __ctalkSelfPrintOn (void) {

  char fmtbuf[MAXMSG];
  OBJECT *fmt_value_obj;
  int retval = 0; /* Avoid a warning. */
  STDARG_CALL_INFO stdarg_call_info;
  OBJECT *arg0_obj = NULL;        /* Avoid a warning. */
  int k;
  OBJECT *read_value_obj;
  METHOD *method;
  
  method = __ctalkRtGetMethod ();

#if ((defined(__linux__) && defined(__i386__)) || \
      defined (__CYGWIN__) || defined (__DJGPP__) || \
     (defined(__sparc__) && defined(__svr4__)))

  if (!strcmp (method -> rcvr_class_obj -> __o_name, STRING_CLASSNAME) ||
      __ctalkIsSubClassOf (method -> rcvr_class_obj -> __o_name,
			   STRING_CLASSNAME)) {
    stdarg_call_info.libcfn = (int (*)())sprintf;
  } else {
    if (!strcmp (method -> rcvr_class_obj -> __o_name, 
		 "WriteFileStream") ||
	__ctalkIsSubClassOf (method -> rcvr_class_obj -> __o_name,
			     "WriteFileStream")) {
      stdarg_call_info.libcfn = (int (*)())fprintf;
    } else {
      _warning 
	("__ctalkSelfPrintOn: Unimplemented receiver class, \"%s.\"\n",
	 method -> rcvr_class_obj -> __o_name);
      return ERROR;
    }
  }

  arg0_obj = __ctalk_receiver_pop ();
  __ctalk_receiver_push (arg0_obj);
  stdarg_call_info.fmt_arg_n = 0;

  if ((fmt_value_obj = 
       __ctalkGetInstanceVariable (method->args[stdarg_call_info.fmt_arg_n], 
				   "value", FALSE))
      == NULL)
    fmt_value_obj = method -> args[stdarg_call_info.fmt_arg_n];

  if (strcmp (fmt_value_obj -> __o_classname, STRING_CLASSNAME)) {
    _warning ("__ctalkLibcFnWithMethodArgs: Argument type mismatch.\n");
    return ERROR;
  }

  if (*fmt_value_obj -> __o_value == '"')
    strcpy (fmtbuf, TRIM_LITERAL(fmt_value_obj -> __o_value));
  else
    strcpy (fmtbuf, fmt_value_obj -> __o_value);

  tokenize_fmt (fmtbuf);

  for (k = stdarg_call_info.fmt_arg_n + 1, ptr_arg_idx = 0; 
       method->args[k]; k++) {
    if ((read_value_obj = 
	 __ctalkGetInstanceVariable (method->args[k], 
				     "value", FALSE)) == NULL)
      read_value_obj = method -> args[k];
    strcpy (ptr_args[ptr_arg_idx++], read_value_obj -> __o_value);
  }

  retval = __call_printon_fn_w_args (arg0_obj, fmtbuf, method, &stdarg_call_info);

  /*
   *  All functions in the C library that use stdargs return int.
   */

  return retval;

#else /* #if defined(__linux__) && defined(__i386__) ...*/

  _warning ("Functions that use stdargs are not implemented for this system.\n");
  return NULL;
#endif
}

int __call_printon_fn_w_args (OBJECT *arg0_obj, char *fmt, METHOD *method, 
			      STDARG_CALL_INFO *stdarg_call_info) {

  int i, j, k, retval;
  int fmtsize;
  int bufsize = MAXMSG;
  char *arg0_buf;

  retval = 0;

  arg0_buf = __xalloc (bufsize + 1);
  fmtsize = strlen (fmt);

  if (stdarg_call_info -> libcfn == (int (*)())sprintf) {
    *arg0_buf = '\0';
    for (i = 0, j = 0, k = 1; (i <= fmt_tok_idx) && fmt_tokens[i]; i++) {
      if (strchr (fmt_tokens[i], '%')) {
	retval += sprintf (tmparg0buf, 
			   __scalar_fmt_conv (fmt_tokens[i], 
					      ptr_args[j++],
					      method -> args[k++]));
      } else {
	retval += sprintf (tmparg0buf, fmt_tokens[i]);
      }
      fmtsize += strlen (tmparg0buf);
      if (fmtsize >= bufsize) {
	while (fmtsize > bufsize)
	  bufsize *= 2;
	arg0_buf = realloc (arg0_buf, bufsize);
      }
      strcat (arg0_buf, tmparg0buf);
    }
    __ctalkSetObjectValueVar(arg0_obj, arg0_buf);
    free (arg0_buf);
    return retval;
  }

  if (stdarg_call_info -> libcfn == (int (*)())fprintf) {
    OBJECT *arg0_val_obj;
    FILE *f_arg0;
    if ((arg0_val_obj = 
	 __ctalkGetInstanceVariable (arg0_obj, "value", FALSE)) != NULL) {
      sscanf (arg0_val_obj -> __o_value, "0x%p", &f_arg0);
    } else {
      sscanf (arg0_obj -> __o_value, "0x%p", &f_arg0);
    }
    for (i = 0, j = 0, k = 1; (i <= fmt_tok_idx) && fmt_tokens[i]; i++) {
      if (strchr (fmt_tokens[i], '%')) {
 	retval += fprintf (f_arg0,
			   __scalar_fmt_conv 
			   (fmt_tokens[i], 
			    ptr_args[j++],
			    method -> args[k++]));
       } else {
	 retval += fprintf (f_arg0, fmt_tokens[i]);
       }
     }
     return retval;
   }

  return ERROR;

}

